/*
 * Copyright (c) 2007-2009  DMC Co., Ltd.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 */
/*
 * EUC
 * tab=4
 */
#if KER26
#define __KERNEL__	1
#if 0
#include <linux/config.h>
#endif
#include <linux/module.h>
#include <linux/init.h>
#include <linux/sched.h>
#include <linux/delay.h>
#include <linux/errno.h>
#include <linux/interrupt.h>
#include <linux/ioport.h>
#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/pci.h>
#include <linux/pnp.h>
#include <linux/sysctl.h>
#include <linux/poll.h>
#include <asm/io.h>
#if 0
#include <asm/dma.h>
#include <asm/uaccess.h>
#include <asm/unistd.h>
#endif
#else
#include <linux/module.h>
#include <linux/kernel.h> // printk
//#include <linux/sched.h> // current->com, current->pid
#include <asm/io.h> // inb,outb
#include <linux/init.h> // module_init,module_exit
#include <asm/uaccess.h> //  copy_to_user, copy_from_user
#include <linux/ioport.h> // check_region,release_region
#include <linux/errno.h>
#include <linux/delay.h>
#include <linux/time.h>
#include <linux/poll.h>
#endif

//#define TEST_A		1

#include "usbdmc.h"
#include "../vers.h"

#if 0
#define DEBUG_FUNC_ENTRY
#define DEBUG_HONDA1
#endif

#define	DEVICE_NAME				"tpdmc"
static char *tpdmc_version 		= FILTERUSB_VERSION;
#if KER26
static const char *RAW_DEVICE_FMT 	= "/dev/dmc%d";
#else
static const char *RAW_DEVICE_FMT 	= "/dev/usb/dmc%d";
#endif
typedef struct tpdmc
{
	int noblk_mode;
	int is_open;
	struct file *fd;
	int reinit_req;
	int inum;
#if TEST_A
	int testcount;
#endif
} tpdmcrec;

#define	MAXIF	2
static tpdmcrec *rawdev = NULL;
static int tpdmc_major  = 200;
#define	TPDMC_MINOR		1

static int tpdmc_open(struct inode *inode, struct file *file);
static int tpdmc_close(struct inode *inode, struct file *file);
#if HAVE_UNLOCKED_IOCTL
static int tpdmc_ioctl(struct file *file,unsigned int cmd, unsigned long arg);
#else
static int tpdmc_ioctl(struct inode *inode, struct file *file,
               unsigned int cmd, unsigned long arg);
#endif
static ssize_t tpdmc_read(struct file * file, char * buffer, size_t count, loff_t *ppos);
static unsigned int tpdmc_poll(struct file *file, poll_table * wwait);
static void tpdmc_term(void);
static int tpdmc_init(void);
static int tpdmc_rawdevopen(int first,int inum);
static void tpcdmc_rawdevclose(tpdmcrec *p);

#define	open(a,b,c)		filp_open(a,b,c)
#define	close(a)		filp_close(a,0)
#define	read(a,b,c,d)	a->f_op->read(a,b,c,d)
#define	poll(a,b)		a->f_op->poll(a,b)
#if HAVE_UNLOCKED_IOCTL
#define	ioctl(a,b,c)	a->f_op->unlocked_ioctl(a,b,c)
#else
#define	ioctl(a,b,c)	a->f_op->ioctl(a->f_dentry->d_inode,a,b,c)
#endif

/************************************************************************
 * open
 */
static int tpdmc_open(struct inode *inode, struct file *pfile)
{
int res;
int inum;
#if KER26
	inum = iminor(inode) - TPDMC_MINOR;
#else
	inum = MINOR(inode->i_rdev) - TPDMC_MINOR;
#endif
	//MOD_INC_USE_COUNT;
#if defined(DEBUG_FUNC_ENTRY)
	printk(KERN_INFO "tpdmc_open entry\n");
#endif
	rawdev[inum].noblk_mode = pfile->f_flags & O_NONBLOCK ? 1 : 0;
	res = tpdmc_rawdevopen(1,inum);
	rawdev[inum].is_open = 1;
	rawdev[inum].reinit_req = 0;
	rawdev[inum].inum = inum;
	pfile->private_data = &rawdev[inum];
	return(0);
}
static int tpdmc_rawdevopen(int first,int inum)
{
struct file *fd;
int md;
char s[100];
	md = rawdev[inum].noblk_mode ? O_NONBLOCK : 0;
	sprintf(s,RAW_DEVICE_FMT,inum);
	fd = open(s,O_RDWR|md,0600);
	if( IS_ERR(fd) || fd == NULL )
	{
		if( first )
			printk(KERN_INFO  "tpdmc_open/(%s) err $%x\n",s,(int)fd);
		return(-1);
	}
	rawdev[inum].fd = fd;
	//printk(KERN_INFO  "tpdmc_rawdevopen/ok $%x \n",(int)rawdev[inum].fd);
	return(0);
}
/************************************************************************
 * close
 */
static int tpdmc_close(struct inode *inode, struct file *pfile)
{
tpdmcrec *p;
#if defined(DEBUG_FUNC_ENTRY)
	printk("tpdmc_close entry\n");
#endif
	p = pfile->private_data;
	tpcdmc_rawdevclose(p);
	p->is_open = 0;
	return(0);
}
static void tpcdmc_rawdevclose(tpdmcrec *p)
{
#if defined(DEBUG_FUNC_ENTRY)
	printk(KERN_INFO "tpdmc_rawdevclose entry %d %x\n",p->is_open,p->fd);
#endif
	if( p->fd )
	{
		close(p->fd);
		p->fd = NULL;
	}
	//printk(KERN_INFO "tpdmc_rawdevclose ok\n");
}
/************************************************************************
 * read
 */
static ssize_t tpdmc_read(struct file *pfile, char *cbuff, size_t size, loff_t *ppos)
{
int res;
tpdmcrec *p;
int inum;
#if 0	//defined(DEBUG_FUNC_ENTRY)
	printk("tpdmc_read entry\n");
#endif
	p = pfile->private_data;
	inum = p->inum;
	if( p->fd == NULL )
	{
		if( tpdmc_rawdevopen(0,inum) != 0 )
			return(0);
		p->reinit_req = 1;
		//printk("tpdmc_read reinit ok\n");
	}
	if( p->reinit_req )
	{
		p->reinit_req = 0;
		cbuff[0] = 0x12;
		return(1);
	}
	res = read(p->fd,cbuff,size,ppos);
	//if( res != -EAGAIN )
	//	printk("tpdmc_read %d\n",res);
	if( res < 0 )
	{
		if( res != -EAGAIN )
			tpcdmc_rawdevclose(p);
		res = 0;
	}
	return(res);
}
/************************************************************************
 * poll
 */
static u_int tpdmc_poll(struct file *file, poll_table * wwait)
{
int res;
tpdmcrec *p;
int inum;
	//printk(KERN_INFO  "tpdmc_poll\n");
	p = file->private_data;
	inum = p->inum;
	if( p->fd == NULL )
	{
		if( tpdmc_rawdevopen(0,inum) != 0 )
			return(0);
		p->reinit_req = 1;
		return(POLLIN|POLLRDNORM);
	}
	res = poll(p->fd,wwait);
	if( res < 0 )
	{
		//printk(KERN_INFO  "tpdmc_poll res=%x\n",res);
		tpcdmc_rawdevclose(p);
		res = 0;
	}
	return(res);
}
/************************************************************************
 * ioctl
 */
#if HAVE_UNLOCKED_IOCTL
static int tpdmc_ioctl(struct file*pfile, unsigned int cmd, unsigned long arg)
#else
static int tpdmc_ioctl(struct inode *inode, struct file*pfile, unsigned int cmd, unsigned long arg)
#endif
{
int res;
tpdmcrec *p;
int inum;
#if defined(DEBUG_FUNC_ENTRY)
	printk("tpdmc_ioctl entry. cmd=%d \n",cmd);
#endif
	p = pfile->private_data;
	inum = p->inum;
	if( p->fd == NULL )
	{
		if( tpdmc_rawdevopen(0,inum) != 0 )
			return(-ENOTCONN);
	}
    switch(cmd)
	{
    case USBDMC_RESET:
		res = ioctl(p->fd,USBDMC_RESET,0);
#if TEST_A
		p->testcount++;
		printk(KERN_INFO "%s testcount=%d\n",__FUNCTION__,p->testcount);
		if( p->testcount >= 3 && p->testcount <= 6 )
		{
			res = -EIO;
		}
#endif
		break;
    case USBDMC_SPECIFY:
		res = ioctl(p->fd,USBDMC_SPECIFY,(u_long)arg);
		break;
    case USBDMC_GOTO_XYP:
		res = ioctl(p->fd,USBDMC_GOTO_XYP,0);
		break;
    case USBDMC_GOTO_IDLE:
		res = ioctl(p->fd,USBDMC_GOTO_IDLE,0);
		break;
    case USBDMC_SETUP:
		res = ioctl(p->fd,USBDMC_SETUP,0);
		break;
    case USBDMC_READ_EEPROM:
		res = ioctl(p->fd,USBDMC_READ_EEPROM,(u_long)arg);
		break;
    case USBDMC_WRITE_EEPROM:
		res = ioctl(p->fd,USBDMC_WRITE_EEPROM,(u_long)arg);
		break;
    case USBDMC_READ_STAT:
		res = ioctl(p->fd,USBDMC_READ_STAT,(u_long)arg);
		break;
    case USBDMC_WRITE_MEM:
		res = ioctl(p->fd,USBDMC_WRITE_MEM,(int *)arg);
		break;
    case USBDMC_READ_MEM:
		res = ioctl(p->fd,USBDMC_READ_MEM,(int *)arg);
		break;
    case USBDMC_SET_OFFDELAY:
		res = ioctl(p->fd,USBDMC_SET_OFFDELAY,(int *)arg);
		break;
    case USBDMC_GET_OFFDELAY:
		res = ioctl(p->fd,USBDMC_GET_OFFDELAY,(int *)arg);
		break;
    default:
		return(-ENOIOCTLCMD);
    }
#if 0
	printk(KERN_INFO  "tpdmc_ioctl cmd=%d res=%d\n",cmd,res);
#endif
    switch(cmd)
	{
    case USBDMC_READ_EEPROM:
    case USBDMC_WRITE_EEPROM:
		break;
	default:
		if( res < 0 )
			tpcdmc_rawdevclose(p);
		break;
	}
	return(res);
}
/************************************************************************
 * interface structure
 */
#if KER26
static struct file_operations ops = {
	.open    = tpdmc_open,
#if HAVE_UNLOCKED_IOCTL
	.unlocked_ioctl   = tpdmc_ioctl,
#else
	.ioctl   = tpdmc_ioctl,
#endif
	.read    = tpdmc_read,
	.poll    = tpdmc_poll,
	.release = tpdmc_close
};
#else
static struct file_operations ops = {
	open    : tpdmc_open,
	ioctl   : tpdmc_ioctl,
	read    : tpdmc_read,
	poll    : tpdmc_poll,
	release : tpdmc_close,
};
#endif

#ifdef MODULE
/************************************************************************
 * tpdmc_term
 */
static void tpdmc_term(void)
{
#if defined(DEBUG_FUNC_ENTRY)
	printk("tpdmc_term entry\n");
#endif
	if( rawdev[0].is_open == 0 
	 && rawdev[1].is_open == 0 )
	{
		unregister_chrdev(tpdmc_major,DEVICE_NAME);
		kfree(rawdev);
	}
	printk("tpdmc_term/success\n");
}
/************************************************************************
 * tpdmc_init
 */
static int tpdmc_init(void)
{
int err;
#if defined(DEBUG_FUNC_ENTRY)
	printk("tpdmc_init entry\n");
#endif
	err = register_chrdev(tpdmc_major, DEVICE_NAME, &ops);
	if(err != 0)
	{
		printk("tpdmc_init/register_chrawdev error\n");
		return(-ENODEV);
	}
	if( !rawdev )
	{
    	if( (rawdev = kmalloc(sizeof(tpdmcrec)*MAXIF, GFP_KERNEL)) == NULL )
		{
        	printk(KERN_ERR "tpdmc_init/no memory\n");
        	return 0;
    	}
		memset(rawdev,0,sizeof(tpdmcrec)*MAXIF);
	}
#if 0
	rawdev->is_open = 0;
	rawdev->fd = NULL;
#endif
	printk("DMC Touch Panel Filter Driver Version %s Major=%d\n"
				,tpdmc_version,tpdmc_major);
	return 0;
}
module_init(tpdmc_init)
module_exit(tpdmc_term)
MODULE_DESCRIPTION("DMC Touch Panel Filter Driver");
MODULE_AUTHOR("DMC");
MODULE_LICENSE("GPL");

#endif

